#include <stdio.h>
#include <sys/types.h>
#include <math.h>
#include <cm/cmmd.h>
#include "const.h"


main (argc,argv)
int argc;
char **argv;
{ 
  int i,j,k,nproc;
  int Arows,Acols,Brows,Bcols;
  int num_of_Acols_per_proc, elem_A_size;
  int num_of_Ccols_per_proc, elem_C_size;
  FILE *fd, *fdout,*fopen();
  float *row_A_arr, *row_C_arr;

  printf("\n Executing host program... (pid=%d)\n", getpid());
  if (argc < 4) {
      printf("Usage A_rows A_cols B_rows B_cols A_file_name B_filename C_file_name\n");
      exit(0);
  }
  sscanf(argv[1],"%d",&Arows);
  sscanf(argv[2],"%d",&Acols);
  sscanf(argv[3],"%d",&Brows);
  sscanf(argv[4],"%d",&Bcols);
  if (Brows != Acols) {
     printf("B_rows not equal to A_cols\n");
     exit(1);
  }
  nproc = CMMD_partition_size();
  if ((Arows%nproc) != 0) {
    printf("# of Arows not multiple of # of processors\n");
    exit(1);
  }
  if ((Bcols%nproc) != 0) {
    printf("# of Bcols not multiple of # of processors\n");
    exit(1);
  }
  CMMD_enable(); 	     

  /* Start execution of node programs and synchronize with each other
     and host */

  printf("\n Executing node programs...");
  node_program();
  CMMD_sync_host_with_nodes();    /* Global sync. to synchronize clocks */
  CMMD_bc_from_host(&Arows,sizeof(int));
  CMMD_bc_from_host(&Acols,sizeof(int));
  CMMD_bc_from_host(&Brows,sizeof(int));
  CMMD_bc_from_host(&Bcols,sizeof(int));

  distribute_cols_uniformly(Brows,Bcols,argv[6]);

  row_A_arr = (float *) calloc(Acols, sizeof(float));
  row_C_arr = (float *) calloc(Bcols, sizeof(float));
  fd = fopen(argv[5],"r");    /* open the file */
  fdout = fopen(argv[7],"w");
  /* Now read each row from file and use CMMD_distrib_to_nodes to distribute the data */
  num_of_Acols_per_proc = Acols/CMMD_partition_size();
  elem_A_size = Acols*sizeof(float);
  num_of_Ccols_per_proc = Bcols/CMMD_partition_size();
  elem_C_size = num_of_Ccols_per_proc*sizeof(float);
  for (i=0; i<Arows; i++) {
        for (j=0; j<Acols; j++) {
           fscanf(fd,"%f",&row_A_arr[j]);
        }
        CMMD_bc_from_host(row_A_arr,elem_A_size);
        CMMD_gather_from_nodes(row_C_arr,elem_C_size);
        for(j=0; j<Bcols; j++)
            fprintf(fdout,"%f  ",row_C_arr[j]);
        fprintf(fdout,"\n");
  }
  fclose(fd);
  fclose(fdout);
  CMMD_disable();
}



distribute_cols_uniformly(nrows,ncols,fname)
int nrows,ncols;
char *fname;
{
    int i,j,num_of_elems_per_proc,elem_size;
    FILE *fd, *fopen();
    float *array;

    /* first create an array of size num_of_cols which is the size of one row. */
    array = (float *) calloc(ncols, sizeof(float));
    if (array == NULL) {
       printf("can't allocate memory\n");
       exit(1);
    }
    /* open the file */
    fd = fopen(fname,"r");
    if (fd == NULL) {
       printf("can't open the file %s\n",fname);
       exit(1);
    }
  /* Now read each row from file and use CMMD_distrib_to_nodes to distribute the data */
    num_of_elems_per_proc = ncols/CMMD_partition_size();
    elem_size = num_of_elems_per_proc*sizeof(float);
    for (i=0; i<nrows; i++) {
        for (j=0; j<ncols; j++) {
           fscanf(fd,"%f",&array[j]);
        }
        CMMD_distrib_to_nodes(array,elem_size);
    }
    fclose(fd);
}
