#include <stdio.h>
#include <sys/types.h>
#include <math.h>
#include <cm/cmmd.h>
#include "const.h"


void CMPE_node_program()
{ 
  int     i,j,k,size;
  int     Arows,Acols,Brows,Bcols;
  float   **mtxB,**mtxA,**mtxC, **create_matrix();
  int     nextaddr,prvaddr,myaddr;
  int     nproc, hostaddr;
  int     Bcols_per_proc, Arows_per_proc;

  CMMD_sync_with_nodes();
  CMMD_sync_with_host();

  CMMD_receive_bc_from_host(&Arows, sizeof(int));
  CMMD_receive_bc_from_host(&Acols, sizeof(int));
  CMMD_receive_bc_from_host(&Brows, sizeof(int));
  CMMD_receive_bc_from_host(&Bcols, sizeof(int));

  nproc = CMMD_partition_size();
  myaddr = CMMD_self_address();
  nextaddr = (myaddr + 1)%nproc;
  if (myaddr == 0)
     prvaddr = nproc - 1;
  else 
     prvaddr =  myaddr - 1;
  hostaddr = CMMD_host_node();
  /* allocate memory for set of rows and cols of mtxA, mtxB, mtxC to
     received by the node  */
  Bcols_per_proc = Bcols/nproc;
  mtxB = create_matrix(Brows,Bcols_per_proc);
  Arows_per_proc = Arows/nproc;
  mtxA = create_matrix(Arows_per_proc,Acols);
  mtxC = create_matrix(Arows_per_proc,Bcols_per_proc);

  receive_columns_uniformly(Brows,Bcols_per_proc,mtxB);
  /* receive rows of matrix A */
  size = Arows_per_proc*Acols*sizeof(float);
  CMMD_receive(hostaddr,DATA,mtxA[0],size);

  do_multiplication(Arows_per_proc,mtxA,Brows,Bcols_per_proc,mtxB,mtxC);
  for (i=0; i<(nproc-1); i++) {
     if ((myaddr%2) == 0) {
        CMMD_send(prvaddr,WORK,mtxA[0],size);
     }
     else {
        CMMD_receive(nextaddr,WORK,mtxA[0],size);
    }
     if ((myaddr%2) == 0) {
        CMMD_receive(nextaddr,WORK,mtxA[0],size);
     }
     else {
        CMMD_send(prvaddr,WORK,mtxA[0],size);
    }
    do_multiplication(Arows_per_proc,mtxA,Brows,Bcols_per_proc,mtxB,mtxC);
  }
}


do_multiplication(Arows_per_proc,mtxA,Brows,Bcols_per_proc,mtxB,mtxC)
int Arows_per_proc,Brows,Bcols_per_proc;
float **mtxA, **mtxB, **mtxC;
{
   int hostaddr,i,j,k,size;

   size = Bcols_per_proc*sizeof(float);
   hostaddr = CMMD_host_node();
   for (i=0; i<Arows_per_proc; i++) {
        for (j=0; j<Bcols_per_proc; j++) {
            mtxC[i][j] = 0.0;
            for (k=0; k<Brows; k++)
                mtxC[i][j] = mtxC[i][j] + mtxA[i][k]*mtxB[k][j];
        }
        CMMD_send(hostaddr,RESULT,mtxC[i],size);
   }
}




receive_columns_uniformly(nrows,ncols,mtx)
int nrows,ncols;
float **mtx;
{
      int i,length;
      
      length = ncols*sizeof(float);
      for (i=0; i<nrows; i++) {
         CMMD_receive_element_from_host(mtx[i],length);
      }
}

