#include <stdio.h>
#include <sys/types.h>
#include <math.h>
#include <cm/cmmd.h>
#include "const.h"


main (argc,argv)
int argc;
char **argv;
{ 
  int i,j,k,nproc;
  int Arows,Acols,Brows,Bcols;
  int num_of_rows_per_proc,size;
  int num_of_cols_per_proc,src;
  int start_row, start_col;
  FILE *fd, *fopen();
  float **mtxA, **mtxB, **mtxC, **create_matrix();

  printf("\n Executing host program... (pid=%d)\n", getpid());
  if (argc < 4) {
      printf("Usage A_rows A_cols B_rows B_cols A_file_name B_filename C_file_name\n");
      exit(0);
  }
  sscanf(argv[1],"%d",&Arows);
  sscanf(argv[2],"%d",&Acols);
  sscanf(argv[3],"%d",&Brows);
  sscanf(argv[4],"%d",&Bcols);
  if (Brows != Acols) {
     printf("B_rows not equal to A_cols\n");
     exit(1);
  }
  nproc = CMMD_partition_size();
  if ((Arows%nproc) != 0) {
    printf("# of Arows not multiple of # of processors\n");
    exit(1);
  }
  if ((Bcols%nproc) != 0) {
    printf("# of Bcols not multiple of # of processors\n");
    exit(1);
  }
  CMMD_enable(); 	     

  /* Start execution of node programs and synchronize with each other
     and host */

  printf("\n Executing node programs...");
  node_program();
  CMMD_sync_host_with_nodes();    /* Global sync. to synchronize clocks */
  CMMD_bc_from_host(&Arows,sizeof(int));
  CMMD_bc_from_host(&Acols,sizeof(int));
  CMMD_bc_from_host(&Brows,sizeof(int));
  CMMD_bc_from_host(&Bcols,sizeof(int));

  num_of_rows_per_proc = Arows/nproc;
  num_of_cols_per_proc = Bcols/nproc;

  distribute_cols_uniformly(Brows,Bcols,argv[6]);

  mtxA = create_matrix(num_of_rows_per_proc,Acols);
  distribute_rows_uniformly(mtxA,Arows,Acols,argv[5]);

  /* wait for the results to arrive from the processors in the order they
     were sent */
  mtxC = create_matrix(Arows,Bcols);
  size = num_of_cols_per_proc*sizeof(float);
  for (i=0; i<nproc; i++) {
     for (j=i; j<(i+nproc); j++) {
         src = j%nproc;
         start_row = src*num_of_rows_per_proc;
         start_col = (j-i)*num_of_cols_per_proc;
         for (k=start_row; k<(start_row+num_of_rows_per_proc); k++) {
             CMMD_receive(src,RESULT, &mtxC[k][start_col], size);
         }
     }
  }
  CMMD_disable();
  fd = fopen(argv[7],"w");
  if (fd == NULL) {
     printf("can't open the file %s\n",argv[7]);
     exit(1);
  }
  for (i=0; i<Arows; i++) {
    for(j=0; j<Bcols; j++)
       fprintf(fd,"%f  ",mtxC[i][j]);
    fprintf(fd,"\n");
  }
  fclose(fd);
}



distribute_cols_uniformly(nrows,ncols,fname)
int nrows,ncols;
char *fname;
{
    int i,j,num_of_elems_per_proc,elem_size;
    FILE *fd, *fopen();
    float *array;

    /* first create an array of size num_of_cols which is the size of one row. */
    array = (float *) calloc(ncols, sizeof(float));
    if (array == NULL) {
       printf("can't allocate memory\n");
       exit(1);
    }
    /* open the file */
    fd = fopen(fname,"r");
    if (fd == NULL) {
       printf("can't open the file %s\n",fname);
       exit(1);
    }
  /* Now read each row from file and use CMMD_distrib_to_nodes to distribute the data */
    num_of_elems_per_proc = ncols/CMMD_partition_size();
    elem_size = num_of_elems_per_proc*sizeof(float);
    for (i=0; i<nrows; i++) {
        for (j=0; j<ncols; j++) {
           fscanf(fd,"%f",&array[j]);
        }
        CMMD_distrib_to_nodes(array,elem_size);
    }
    fclose(fd);
}



distribute_rows_uniformly(mtx,nrows,ncols,fname)
float **mtx;
int nrows,ncols;
char *fname;
{
    int i,j,k,nproc;
    int num_of_rows_per_proc,size;
    FILE *fd, *fopen();
    float *array;

    /* open the file */
    fd = fopen(fname,"r");
    if (fd == NULL) {
       printf("can't open the file %s\n",fname);
       exit(1);
    }
  /* Now read each row from file and use CMMD_distrib_to_nodes to distribute the data */
    nproc = CMMD_partition_size();
    num_of_rows_per_proc = nrows/nproc;
    size = num_of_rows_per_proc*ncols*sizeof(float);
    for (i=0; i<nproc; i++) {
       for (j=0; j<num_of_rows_per_proc; j++) {
           for (k=0; k<ncols; k++) 
              fscanf(fd,"%f",&(mtx[j][k]));
        }
        CMMD_send(i,DATA,mtx[0],size);
    }
    fclose(fd);
}
